# Aspidites is Copyright 2021, Ross J. Duff.
# See LICENSE.txt for more info.

from typing import Callable, Tuple

from pyrsistent import pmap

from Aspidites._vendor.contracts import check, contract, new_contract


@new_contract
@contract
def heritable(bases: 'tuple') -> 'bool':
    """Convenience clause for a tuple of heritable bases"""
    return not bool(len(tuple(b for b in bases if isinstance(b, final))))


# noinspection PyPep8Naming
class final(type):
    def __new__(mcs, name, bases, classdict):
        check('heritable', bases)
        return type.__new__(mcs, name, bases, dict(classdict))



import unittest
from enum import Enum
import re

import Aspidites.monads
from Aspidites._vendor.pampy import ANY, match_value, match, HEAD, TAIL, MatchError


class PampyBasicTests(unittest.TestCase):

    def test_match_value(self):
        def match_value_bool(pat, var):
            return match_value(pat, var)[0]

        self.assertTrue(match_value_bool(3, 3))
        self.assertTrue(match_value_bool(int, 3))
        self.assertTrue(match_value_bool("ok", "ok"))
        self.assertTrue(match_value_bool(str, "ok"))
        self.assertTrue(match_value_bool(ANY, "ok"))
        self.assertTrue(match_value_bool(ANY, 3))
        self.assertTrue(match_value_bool(ANY, 3.3))
        self.assertTrue(match_value_bool(True, True))
        self.assertTrue(match_value_bool(False, False))
        self.assertFalse(match_value_bool(1, 3))

    def test_match_value_None(self):
        self.assertEqual(match_value(None, None), (True, []))

    def test_match_None(self):
        self.assertEqual(match(None, None, 'None', ANY, 'else'), 'None')

    def test_equality_comparison_should_be_strict(self):
        self.assertEqual(match_value(1, True), (False, []))
        self.assertEqual(match_value(0, False), (False, []))
        self.assertEqual(match_value(1.0, 1), (False, []))
        self.assertEqual(match_value(0.0, 0), (False, []))
        self.assertEqual(match_value(1.0, 1), (False, []))

    def test_match_value_is_vs_equal(self):
        a = 'x' * 1000000
        b = 'x' * 1000000
        self.assertEqual(a, b)
        self.assertFalse(a is b)
        self.assertEqual(match_value(a, b), (True, []))
        self.assertEqual(match(a, b, True), True)

    def test_match_value_var_extraction(self):
        self.assertEqual(match_value(3, 3), (True, []))
        self.assertEqual(match_value(ANY, 3), (True, [3]))
        self.assertEqual(match_value(ANY, 'ok'), (True, ['ok']))

    def test_match_value_callable_pattern(self):
        self.assertEqual(match_value(3, lambda: True), (False, []))

        func = lambda x: True
        self.assertEqual(match_value(callable, func), (True, [func]))

    def test_match_mylen(self):
        def mylen(l):
            return match(l,
                         [], 0,
                         [HEAD, TAIL], lambda head, tail: 1 + mylen(tail))

        self.assertEqual(mylen([]), 0)
        self.assertEqual(mylen([1]), 1)
        self.assertEqual(mylen([1, 2, 3]), 3)

    def test_match(self):
        self.assertTrue(match(3, 3, True))
        self.assertTrue(match(3, ANY, True))
        self.assertTrue(match(3,
                              1, False,
                              2, False,
                              ANY, True))

        self.assertTrue(match([1, 2],
                              [1], False,
                              [1, 2], True))

    def test_match_strict_returns_undefined(self):
        assert match(3, 2, True) is Aspidites.monads.Undefined()

    def test_match_not_strict_returns_false(self):
        assert match(3, 2, True, strict=False) is False

    def test_match_default(self):
        self.assertFalse(match(3, 2, True, default=False))
        self.assertEqual(match(3, 2, True, default=6), 6)

    def test_match_arguments_passing(self):
        self.assertEqual(match([1, 2, 3], [1, ANY, 3], lambda x: x), 2)
        self.assertEqual(match([1, 2, 3], [1, 2, 3], lambda x: x), [1, 2, 3])

    def test_match_action_can_be_empty_list(self):
        self.assertEqual(match(True, True, []), [])

    # TODO: replace with Undefined test
    # def test_match_raise_lambda_error(self):
    #     with self.assertRaises(MatchError) as err:
    #         match([1, 2, 3], [1, ANY, 3], lambda x: "xxxxx {}".format(x))
    #     self.assertIn('lambda', str(err.exception))
    #     self.assertIn('xxxxx', str(err.exception))
    #     # print(err.exception)

    def test_match_class_hierarchy(self):
        class Pet: pass
        class Dog(Pet): pass
        class Cat(Pet): pass
        class Hamster(Pet): pass

        def what_is(x):
            return match(x,
                Dog, 'dog',
                Cat, 'cat',
                Pet, 'any other pet',
                ANY, 'this is not a pet at all',
            )

        self.assertEqual(what_is(Cat()), 'cat')
        self.assertEqual(what_is(Dog()), 'dog')
        self.assertEqual(what_is(Hamster()), 'any other pet')
        self.assertEqual(what_is(Pet()), 'any other pet')
        self.assertEqual(what_is(True), 'this is not a pet at all')

    def test_regex_groups(self):
        def what_is(pet):
            return match(pet,
                re.compile(r'(\w+)-(\w+)-cat$'),     lambda name, my: 'cat '+name,
                re.compile(r'(\w+)-(\w+)-dog$'),     lambda name, my: 'dog '+name,
                ANY,                                  "something else"
            )

        self.assertEqual(what_is('fuffy-my-dog'), 'dog fuffy')
        self.assertEqual(what_is('puffy-her-dog'), 'dog puffy')
        self.assertEqual(what_is('carla-your-cat'), 'cat carla')
        self.assertEqual(what_is('roger-my-hamster'), 'something else')

    def test_regex_no_group(self):
        def what_is(pet):
            return match(pet,
                re.compile('fuffy-cat$'), lambda x: 'fuffy-cat',
                ANY, "something else"
            )

        self.assertEqual(what_is('my-fuffy-cat'), 'fuffy-cat')

    def test_match_enum(self):
        class Color(Enum):
            RED = 1
            GREEN = 2
            BLUE = 3

        self.assertEqual(match(Color.RED, Color.BLUE, "blue", Color.RED, "red", ANY, "else"), "red")
        self.assertEqual(match(Color.RED, Color.BLUE, "blue", Color.GREEN, "green", ANY, "else"), "else")
        self.assertEqual(match(1, Color.BLUE, "blue", Color.GREEN, "green", ANY, "else"), "else")



import unittest
from datetime import datetime
from typing import (
    Union,
    Optional,
    Tuple,
    List,
    NewType,
)

from functools import reduce

from Aspidites._vendor import match, REST, TAIL
from Aspidites._vendor import ANY


class PampyElaborateTests(unittest.TestCase):

    def test_fibonacci(self):
        def fib(n):
            return match(n,
                         1, 1,
                         2, 1,
                         ANY, lambda x: fib(x - 1) + fib(x - 2))

        self.assertEqual(fib(1), 1)
        self.assertEqual(fib(7), 13)

    def test_slide1(self):
        _input = [1, 2, 3]
        pattern = [1, ANY, 3]
        action = lambda x: "it's {}".format(x)
        self.assertEqual(match(_input, pattern, action), "it's 2")

    def test_parser(self):
        def parser(exp):
            return match(exp,
                         3, "the integer 3",
                         float, "any float number",
                         int, "any integer",
                         "ciao", "the string ciao",
                         dict, "any dictionary",
                         str, "any string",
                         (int, int), "a tuple made of two ints",
                         [1], "the list [1]",
                         [1, 2, 3], "the list [1, 2, 3]",
                         [1, ANY, 3], "the list [1, ANY, 3]",
                         (str, str), lambda a, b: "%s %s" % (a, b),
                         [1, 2, ANY], lambda x: "the list [1, 2, ANY]",
                         [1, 2, 4], "the list [1, 2, 4]",  # this can never be matched

                         [1, [2, ANY], ANY], lambda a, b: "[1, [2, %s], %s]" % (a, b),
                         )

        assert parser(3) == "the integer 3"
        assert parser(5) == "any integer"
        assert parser("ciao") == "the string ciao"
        assert parser("x") == "any string"
        assert parser({'a': 1}) == "any dictionary"
        assert parser([1]) == "the list [1]"
        assert parser([1, 2, 3]) == "the list [1, 2, 3]"
        assert parser([1, 2, 4]) == "the list [1, 2, ANY]"
        assert parser([1, 3, 3]) == "the list [1, ANY, 3]"
        assert parser(("hello", "world")) == "hello world"
        assert parser([1, [2, 3], 4]) == "[1, [2, 3], 4]"

    def test_lisp(self):
        # A Lisp interpreter in 5 lines
        def lisp(exp):
            return match(exp,
                         int, lambda x: x,
                         callable, lambda x: x,
                         (callable, REST), lambda f, rest: f(*map(lisp, rest)),
                         tuple, lambda t: list(map(lisp, t)),
                         )

        plus = lambda a, b: a + b
        minus = lambda a, b: a - b
        self.assertEqual(lisp((plus, 1, 2)), 3)
        self.assertEqual(lisp((plus, 1, (minus, 4, 2))), 3)
        self.assertEqual(lisp((reduce, plus, (1, 2, 3))), 6)
        self.assertEqual(lisp((reduce, plus, (range, 10))), 45)

    def test_myzip(self):
        def myzip(a, b):
            return match((a, b),
                         ([], []), [],
                         ([ANY, TAIL], [ANY, TAIL]), lambda ha, ta, hb, tb: [(ha, hb)] + myzip(ta, tb)
                         )

        self.assertEqual(myzip([1, 2, 3], [4, 5, 6]), [(1, 4), (2, 5), (3, 6)])
        self.assertEqual(myzip(range(5), range(5)), list(zip(range(5), range(5))))

    def test_lambda_cond(self):
        cond = lambda x: x < 10
        self.assertEqual(match(3, cond, "action", ANY, "else"), "action")
        self.assertEqual(match(11, cond, "action1", ANY, "else"), "else")

    def test_lambda_cond_arg_passing(self):
        def f(x):
            return match(x,
                         lambda x: x % 2 == 0, lambda x: "even %d" % x,
                         lambda x: x % 2 != 0, lambda x: "odd %d" % x
                         )

        self.assertEqual(f(3), "odd 3")
        self.assertEqual(f(18), "even 18")
        self.assertEqual(f(18), "even 18")

    def test_animals(self):
        pets = [
            {'type': 'dog', 'pet-details': {'name': 'carl', 'cuteness': 4}},
            {'type': 'dog', 'pet-details': {'name': 'john', 'cuteness': 3}},
            {'type': 'cat', 'pet-details': {'name': 'fuffy', 'cuty': 4.6}},
            {'type': 'cat', 'cat-details': {'name': 'bonney', 'cuty': 7}},
        ]

        def avg_cuteness_pampy():
            cutenesses = []
            for pet in pets:
                match(pet,
                      {ANY: {"cuteness": ANY}}, lambda key, x: cutenesses.append(x),
                      {ANY: {"cuty": ANY}}, lambda key, x: cutenesses.append(x)
                      )
            return sum(cutenesses) / len(cutenesses)

        self.assertEqual(avg_cuteness_pampy(), (4 + 3 + 4.6 + 7) / 4)

    def test_advanced_lambda(self):
        def either(pattern1, pattern2):
            """Matches values satisfying pattern1 OR pattern2"""
            def repack(*args):
                return True, list(args)

            def f(var):
                return match(var,
                     pattern1, repack,
                     pattern2, repack,
                     ANY,        (False, [])
                )

            return f

        self.assertEqual(match('str', either(int, str), 'success'), 'success')

        def datetime_p(year: int, month: int, day: int, hour: int = 0, minute: int = 0, second: int = 0):
            """Matches a datetime with these values"""
            def f(var: datetime):
                if not isinstance(var, datetime):
                    return False, []

                args = []
                for pattern, actual in [(year, var.year), (month, var.month), (day, var.day),
                                        (hour, var.hour), (minute, var.minute), (second, var.second)]:
                    if pattern is ANY:
                        args.append(actual)
                    elif pattern != actual:
                        return False, []

                return True, args

            return f

        def test(var):
            return match(var,
                datetime_p(2018, 12, 23), 'full match',
                datetime_p(2018, ANY, ANY), lambda month, day: f'{month}/{day} in 2018',
                datetime_p(ANY, ANY, ANY, ANY, ANY, ANY), 'any datetime',
                ANY, 'not a datetime'
            )

        self.assertEqual(test(datetime(2018, 12, 23)), 'full match')
        self.assertEqual(test(datetime(2018, 1, 2)), '1/2 in 2018')
        self.assertEqual(test(datetime(2017, 1, 2, 3, 4, 5)), 'any datetime')
        self.assertEqual(test(11), 'not a datetime')

    def test_typing_example(self):

        timestamp = NewType("timestamp", Union[float, int])
        year, month, day, hour, minute, second = int, int, int, int, int, int
        day_tuple = Tuple[year, month, day]
        dt_tuple = Tuple[year, month, day, hour, minute, second]

        def datetime_p(patterns: List[str]):
            def f(dt: str):
                for pattern in patterns:
                    try:
                        return True, [datetime.strptime(dt, pattern)]
                    except Exception:
                        continue
                else:
                    return False, []
            return f

        def to_datetime(dt: Union[
            timestamp,
            day_tuple,
            dt_tuple,
            str,
        ]) -> Optional[datetime]:
            return match(dt,
                timestamp, lambda x: datetime.fromtimestamp(x),
                Union[day_tuple, dt_tuple], lambda *x: datetime(*x),
                datetime_p(["%Y-%m-%d", "%Y-%m-%d %H:%M:%S"]), lambda x: x,
                ANY, None
            )

        key_date_tuple = (2018, 1, 1)
        detailed_key_date_tuple = (2018, 1, 1, 12, 5, 6)
        key_date = datetime(*key_date_tuple)
        detailed_key_date = datetime(*detailed_key_date_tuple)

        self.assertEqual(to_datetime(key_date_tuple), key_date)
        self.assertEqual(to_datetime(detailed_key_date_tuple), detailed_key_date)

        key_date_ts = key_date.timestamp()
        detailed_key_date_ts = int(detailed_key_date.timestamp())
        self.assertEqual(to_datetime(key_date_ts), key_date)
        self.assertEqual(to_datetime(detailed_key_date_ts), detailed_key_date)

        key_date_ts_str_a = key_date.strftime("%Y-%m-%d")
        key_date_ts_str_f = key_date.strftime("%Y-%m-%d %H:%M:%S")
        key_date_ts_str_w = key_date.strftime("%m-%Y-%d")
        self.assertEqual(to_datetime(key_date_ts_str_a), key_date)
        self.assertEqual(to_datetime(key_date_ts_str_f), key_date)
        self.assertEqual(to_datetime(key_date_ts_str_w), None)

        detailed_key_date_ts_str = detailed_key_date.strftime("%Y-%m-%d %H:%M:%S")
        self.assertEqual(to_datetime(detailed_key_date_ts_str), detailed_key_date)

        self.assertEqual(to_datetime(set(key_date_tuple)), None)



# Aspidites is Copyright 2021, Ross J. Duff.
# See LICENSE.txt for more info.
import sys
import inspect
import operator as op
import shutil
import sys
import traceback
import typing
from math import inf, isinf
from contextlib import suppress

from _warnings import warn
from pyrsistent import v
from Aspidites import final
from Aspidites._vendor.contracts import ContractNotRespected, contract
from Aspidites._vendor.fn import apply
from Aspidites._vendor.fn.underscore import ArityError

from Aspidites.api import ContractBreachWarning, create_warning


# noinspection PyPep8Naming
def SafeDiv(a, b):
    """IEEE 754-1985 evaluates an expression and replaces indeterminate forms with Undefined instances"""
    if b == 0 or (isinf(a) and isinf(b)):
        stack = inspect.stack()
        w = create_warning(
            stack[0][3],
            [a, b],
            {},
            stack,
            ZeroDivisionError("Division by zero is Undefined; this behavior diverges from IEEE 754-1985.")
        )
        warn(w, category=RuntimeWarning)
        return Undefined(SafeDiv, a=a, b=b)
    return a / b


# noinspection PyPep8Naming
def SafeMod(a, b):
    """IEEE 754-1985 evaluates an expression and replaces indeterminate forms with Undefined instances"""
    if isinf(a) or b == 0:
        stack = inspect.stack()
        w = create_warning(
            stack[0][3],
            [a, b],
            {},
            stack,
            ZeroDivisionError("Modulus by zero is Undefined; this behavior diverges from IEEE 754-1985.")
        )
        warn(w, category=RuntimeWarning)
        return Undefined(SafeMod, a=a, b=b)
    return a % b


# noinspection PyPep8Naming
def SafeExp(a, b):
    if ((a == 0 and b == 0)
            or (isinf(a) and b == 0)
            or (isinf(b) and a == 0)):  # 0**0, inf**0, 0**inf
        stack = inspect.stack()
        w = create_warning(
            stack[0][3], [a, b], {}, stack,
            ArithmeticError("%s**%s" % (str(a), str(b),) + " == Undefined; this behavior diverges from IEEE 754-1985.")
        )
        warn(w, category=RuntimeWarning)
        return Undefined(SafeExp, a=a, b=b)
    try:
        return a**b
    except OverflowError:
        return inf  # just a really big number on most systems


class Maybe:
    """Sandboxes a Surely call and handles ContractNotRespected by returning Undefined"""

    __metaclass__ = final
    __slots__ = v("_func", "_args", "_kwargs", "__instance__")

    def __init__(self, func, *args, **kwargs):
        self._func = func
        self._args = args
        self._kwargs = kwargs
        self.__instance__ = Undefined()

    def __repr__(self):
        return (
            self.__class__.__name__
            + "("
            + ", ".join(
                [
                    str(self._func.__name__),
                    str(self._args).strip("()"),
                    *[str(k) + " = " + str(v) for k, v in self._kwargs.items()],
                ]
            )
            + ")"
            + (
                " -> %s" % Undefined()
                if self.__instance__ == Undefined()
                else " -> %s" % str(self.__instance__)
            )
        )

    def __invert__(self):
        return ~self.__instance__

    def __neg__(self):
        return -self.__instance__

    @property
    def func(self):
        return self._func

    @property
    def args(self):
        return self._args

    @property
    def kwargs(self):
        return self._kwargs

    def __call__(self, warn_undefined=True):
        try:
            with suppress(ValueError):
                val = apply(self.func, self.args, self.kwargs)
            with suppress(UnboundLocalError):
                self.__instance__ = Surely(val)
                # SURELY #
                return self.__instance__
            self.__instance__ = Undefined(self.func, self.args, self.kwargs)
        except (ContractNotRespected, ArityError, ZeroDivisionError) as e:
            if warn_undefined:
                stack = inspect.stack(1)
                w = create_warning(self.func, self.args, self.kwargs, stack, e)
                warn(
                    w,
                    category=ContractBreachWarning
                    if isinstance(e, ContractNotRespected)
                    else RuntimeWarning,
                )
            # UNDEFINED #
            self.__instance__ = Undefined(self.func, self.args, self.kwargs)
            return self.__instance__


class Undefined:
    """A monad for a failed programmatic unit; like NoneType but hashable.
    Falsy singleton"""

    __metaclass__ = final
    __slots__ = v("__weakref__", "__instance__")
    __instance = None

    def __hash__(self):
        return self

    def __eq__(self, other):
        return self.__hash__ == other.__hash__

    def __add__(self, other):
        return other

    def __sub__(self, other):
        return -other

    def __neg__(self):
        return self

    def __invert__(self):
        return self

    def __mul__(self, other):
        return self

    def __truediv__(self, other):
        return self

    def __floordiv__(self, other):
        return self

    def __str__(self):
        return str(self.__repr__())

    def __float__(self):
        return float()

    def __complex__(self):
        return complex()

    def __oct__(self):
        return oct(0)

    def __index__(self):
        return 0

    def __len__(self):
        # Undefined has 0 elements
        return 0

    def __repr__(self):
        return self.__class__.__name__

    def __nonzero__(self):
        return False

    def __call__(self, *args, **kwargs):
        return self.__new__(self.__class__, *args, **kwargs)

    def __new__(mcs, *args, **kwargs):
        if mcs.__instance is None:
            mcs.__instance = super(Undefined, mcs).__new__(mcs, *args, **kwargs)
            mcs.__instance__ = mcs.__instance
        return mcs.__instance__  # instance descriptor from __slots__ -> actual instance


class Surely:
    """A monad for a successful programmatic unit
    Truthy, defers to an instance of a successful computation"""

    __metaclass__ = final
    __slots__ = v(
        "__weakref__", "__instance__" "__str__", "__int__", "__float__", "__complex__"
    )

    # def __try_except_undefined__(self, other=None, call=None):
    #     if call:
    #         # noinspection PyComparisonWithNone
    #         return (
    #             call(self.__instance__)
    #             if other == None
    #             else call(self.__instance__, other)
    #         )
    #     else:
    #         return self.__instance__
    #
    # def __hash__(self):
    #     return self.__try_except_undefined__(call=hash)
    #
    # def __eq__(self, other):
    #     return self.__try_except_undefined__(other, op.eq)
    #
    # def __add__(self, other):
    #     return self.__try_except_undefined__(other, op.add)
    #
    # def __sub__(self, other):
    #     return self.__try_except_undefined__(other, op.sub)
    #
    # def __neg__(self):
    #     return self.__try_except_undefined__(call=op.neg)
    #
    # def __invert__(self):
    #     return self.__try_except_undefined__(call=op.invert)
    #
    # def __mul__(self, other):
    #     return self.__try_except_undefined__(other, op.mul)
    #
    # def __truediv__(self, other):
    #     return self.__try_except_undefined__(other, op.truediv)
    #
    # def __floordiv__(self, other):
    #     return self.__try_except_undefined__(other, op.floordiv)
    #
    # def __oct__(self):
    #     return self.__try_except_undefined__(call=oct)
    #
    # def __nonzero__(self):
    #     return self.__try_except_undefined__(call=bool)
    #
    # @classmethod
    # def __call__(cls, *args, **kwargs):
    #     return cls.__instance__

    def __new__(cls, instance__=Undefined(), *args, **kwargs):
        cls.__instance__ = instance__
        return cls.__instance__



# Aspidites is Copyright 2021, Ross J. Duff.
# See LICENSE.txt for more info.

import os
import py_compile
import sys
from glob import glob
from warnings import warn
import typing as t
from mypy import api
from Aspidites.templates import lib, makefile, pyproject, setup
from pyrsistent import pmap
from hashlib import sha256


def checksum(fname, write=True, check=False):
    base, name = os.path.split(fname)
    fname_sha256 = os.path.join(base, "." + name) + ".sha256"

    def read_sha256(data):
        curr_hash = sha256()
        chunk = data.read(8192)
        while chunk:
            curr_hash.update(chunk)
            chunk = data.read(8192)
        return curr_hash

    if write:
        with open(fname, "rb") as data:
            curr_hash = read_sha256(data)
            with open(fname_sha256, "wb") as digest:
                digest.write(curr_hash.digest())
            return pmap({curr_hash.digest(): fname}).items()[0]  # immutable
    if check:
        with open(fname_sha256, "rb") as digest:
            with open(fname, "rb") as data:
                curr_hash = read_sha256(data)
                old = digest.read()
                new = curr_hash.digest()
                if new == old:
                    print(
                        "sha256 digest check successful: %s, %s == %s"
                        % (fname, new.hex(), old.hex())
                    )
                    return new
                else:
                    print(
                        "sha256 digest failure: %s, %s != %s"
                        % (fname, new.hex(), old.hex())
                    )
                    return ""


class CheckedFileStack:
    def __init__(self, initial=None, pre_size=8192):
        if initial is None:
            initial = {}
        self._files = pmap(initial, pre_size)
        self.all_files = self._files.evolver()

    def register(self, fname):
        self.all_files.set(*checksum(fname))

    def finalize(self):
        return self.all_files.persistent()


_stack = CheckedFileStack()


def _finalize_stack():
    all_file_checksums = _stack.finalize()
    print("running checksums")
    for k, v in all_file_checksums.items():
        digest = checksum(v, write=False, check=True)
        try:
            all_file_checksums.get(digest)
        except AttributeError:
            raise RuntimeError("\nfor file %s\n%s\n  did not match cached digest\n%s")


class CompilerArgSpec(t.TypedDict):

    code: object
    fname: str
    force: bool
    bytecode: bool
    c: bool
    build_requires: t.Union[t.List, str]
    verbose: int


def create_file(fname, mode, stack=_stack, root='', text="# THIS FILE IS GENERATED - DO NOT EDIT #") -> None:

    if len(root) > 0:
        file = os.path.join(root, fname)
    else:
        file = fname
    try:
        open(file, mode).write(text)
    except FileExistsError:
        stack.register(file)


def compile_module(**kwargs):
    code = kwargs['code']
    fname = kwargs['fname']
    force = kwargs['force']
    bytecode = kwargs['bytecode']
    c = kwargs['c']
    build_requires = kwargs['build_requires']
    verbose = kwargs['verbose']
    app_name = os.path.splitext(fname)[0]
    project = os.path.basename(app_name)
    module_name = app_name.replace("/", ".")
    file_c = app_name + ".c"
    root = os.path.dirname(fname)
    mode = "x" if force else "w"
    create_file(fname, mode, root='', text=lib.substitute(code="\n".join(code)))
    create_file('__init__.py', mode=mode, root=root)
    create_file('py.typed', mode=mode, root=root)
    typecheck(module_name)
    create_stubs(fname, app_name)
    create_file('pyproject.toml', mode, root=root,
                text=pyproject.substitute(build_requires=build_requires))
    create_file('Makefile', mode, root=root, text=makefile.substitute(project=project))

    if bytecode:
        fname_pyc = app_name + ".pyc"
        quiet = tuple(reversed(range(3))).index(verbose if verbose < 2 else 2)
        py_compile.compile(fname, fname_pyc, quiet=quiet)
        _stack.register(fname_pyc)

    if c:
        compile_c(fname, force, verbose)
        create_setup(root, **kwargs)
        compile_object(app_name, file_c, root)

    _finalize_stack()


def typecheck(module_name):
    mypy_args = [
        "-m",
        module_name,
        "--follow-imports=skip",
        "--show-error-context",
        "--show-error-codes",
        "--allow-incomplete-defs",
        "--disable-error-code=valid-type",
        "--exclude=builtins"
        # '--disallow-untyped-defs',
        # '--disallow-untyped-calls',
    ]
    print("running mypy", " ".join(mypy_args))
    type_report = None
    error_report = None
    return_code = 0
    try:
        type_report, error_report, return_code = api.run(mypy_args)
    except AttributeError:
        warn("mypy api call failed")

    finally:
        print("mypy type report: ", type_report) if type_report else None
        print("mypy error report: ", error_report) if error_report else None
        print("mypy returned with exit code:", return_code) if return_code else None
        # exit(return_code) if return_code != 0 else print("running compile")


def create_stubs(fname, app_name):
    fname_pyi = app_name + '.pyi'
    stubgen_runner = 'stubgen %s -o .' % (fname)
    print("running %s" % stubgen_runner)
    with os.popen(stubgen_runner) as p:
        print(p.read())
    try:
        _stack.register(fname_pyi)
    except FileNotFoundError as e:
        warn(str(e))
        try:
            print("trying rename %s/__main__.pyi to %s" % (os.getcwd(), fname_pyi))
            os.rename(os.path.join(os.getcwd(), '__main__.pyi'), fname_pyi)
            _stack.register(fname_pyi)
        except FileNotFoundError:
            warn("failed to create stubs")


def compile_c(fname, force, verbose) -> None:
    verb = int(bool(verbose))
    os.popen("cython %s %s %s" % (fname, "--force" * force, "--verbose" * verb))


def create_setup(root, **kwargs) -> None:
    app_name = os.path.splitext(kwargs['fname'])[0]
    module_name = app_name.replace("/", ".")
    create_file('setup.py',
                "x" if kwargs['force'] else "w",
                root=root,
                text=setup.substitute(
                    app_name=module_name,
                    src_file=kwargs['fname'],
                    inc_dirs=[],
                    libs=[],
                    exe_name=app_name,
                    lib_dirs=[],
                    **kwargs))


def compile_object(app_name, file_c, root) -> None:
    glob_so = app_name + ".*.so"
    # TODO: get this working for docker builds
    #  (maybe executable param with os.path.relpath?)
    setup_runner = "%s %s build_ext -b ." % (sys.executable, os.path.join(root, 'setup.py'))
    print("running", setup_runner)
    with os.popen(setup_runner) as p:
        print(p.read())
    _stack.register(file_c)
    for i in glob(glob_so):
        _stack.register(i)



from Aspidites._vendor.RestrictedPython import RestrictingNodeTransformer

import ast


def test_RestrictingNodeTransformer__generic_visit__1():
    """It log an error if there is an unknown ast node visited."""
    class MyFancyNode(ast.AST):
        pass

    transformer = RestrictingNodeTransformer()
    transformer.visit(MyFancyNode())
    assert transformer.errors == [
        'Line None: MyFancyNode statements are not allowed.']
    assert transformer.warnings == [
        'Line None: MyFancyNode statement is not known to RestrictedPython']



import collections.abc as collections


def ist(C):
    def f(x):
        f.__name__ = 'isinstance_of_%s' % C.__name__
        if not isinstance(x, C):
            raise ValueError('Value is not an instance of %s.' % C.__name__)
    return f


def m_new_contract(name, f):
    from Aspidites._vendor.contracts.library.extensions import CheckCallable
    from Aspidites._vendor.contracts.library.extensions import Extension
    Extension.registrar[name] = CheckCallable(f)
    

m_new_contract('Container', ist(collections.Container))
m_new_contract('Iterable', ist(collections.Iterable))

m_new_contract('Hashable', ist(collections.Hashable))
m_new_contract('Generator', ist(collections.Generator))


m_new_contract('Iterator', ist(collections.Iterator))
m_new_contract('Sized', ist(collections.Sized))
m_new_contract('Callable', ist(collections.Callable))
m_new_contract('Sequence', ist(collections.Sequence))
m_new_contract('Set', ist(collections.Set))
m_new_contract('MutableSequence', ist(collections.MutableSequence))
m_new_contract('MutableSet', ist(collections.MutableSet))
m_new_contract('Mapping', ist(collections.Mapping))
m_new_contract('MutableMapping', ist(collections.MutableMapping))
#new_contract('MappingView', ist(collections.MappingView))
#new_contract('ItemsView', ist(collections.ItemsView))
#new_contract('ValuesView', ist(collections.ValuesView))


# Not a lambda to have better messages
def is_None(x): 
    return x is None


m_new_contract('procedure', is_None)
m_new_contract('coroutine', ist(collections.Generator))
m_new_contract('None', is_None)
m_new_contract('NoneType', is_None)




# THIS FILE IS GENERATED - DO NOT EDIT #
from setuptools import setup, Extension
from Cython.Build import cythonize
from Cython.Compiler import Options
Options.annotate = False
Options.annotate_coverage_xml = None
Options.buffer_max_dims = 8
Options.cache_builtins = True
Options.cimport_from_pyx = False
Options.clear_to_none = True
Options.closure_freelist_size = 8
Options.convert_range = True
Options.docstrings = True
Options.embed_pos_in_docstring = False
Options.generate_cleanup_code = False
Options.fast_fail = False
Options.warning_errors = False
Options.error_on_unknown_names = True
Options.error_on_uninitialized = True
Options.gcc_branch_hints = True
Options.lookup_module_cpdef = False
Options.embed = True

exts = [
Extension(
         'Aspidites.woma.library', 
         ['Aspidites/woma/library.pyx'],
         include_dirs=[],
         libraries=[], 
         extra_compile_args=['-Wall'],
         library_dirs=[]
    ),
]

setup(
    name='Aspidites.woma.library',
    ext_modules=cythonize(exts))



import operator
import pytest
from hypothesis import given, strategies as st
from Aspidites._vendor.fn import F, _
from Aspidites._vendor.fn import iters


def test_take():
    assert [0, 1] == list(iters.take(2, range(10)))
    assert [0, 1] == list(iters.take(10, range(2)))


def test_drop():
    assert [3, 4] == list(iters.drop(3, range(5)))
    assert [] == list(iters.drop(10, range(2)))


def test_takelast():
    assert [8, 9] == list(iters.takelast(2, range(10)))
    assert [0, 1] == list(iters.takelast(10, range(2)))


def test_droplast():
    assert [0, 1] == list(iters.droplast(3, range(5)))
    assert [] == list(iters.droplast(10, range(2)))


def test_consume():
    # full consuming, without limitation
    r = range(10)
    assert 10 == len(list(r))
    itr = iter(r)
    iters.consume(itr)
    assert 0 == len(list(itr))


def test_consume_limited():
    r = range(10)
    assert 10 == len(list(r))
    itr = iter(r)
    iters.consume(itr, 5)
    assert 5 == len(list(itr))


def test_nth():
    assert 1 == iters.nth(range(5), 1)
    assert None == iters.nth(range(5), 10)
    assert "X" == iters.nth(range(5), 10, "X")


def test_head():
    assert 0 == iters.head([0, 1, 2])
    assert None == iters.head([])

    def gen():
        yield 1
        yield 2
        yield 3

    assert 1, iters.head(gen())


def test_tail():
    assert [1, 2] == list(iters.tail([0, 1, 2]))
    assert [] == list(iters.tail([]))

    def gen():
        yield 1
        yield 2
        yield 3

    assert [2, 3] == list(iters.tail(gen()))


def test_padnone():
    it = iters.padnone([10, 11])
    assert 10 == next(it)
    assert 11 == next(it)
    assert None == next(it)
    assert None == next(it)


def test_ncycles():
    it = iters.ncycles([10, 11], 2)
    assert 10 == next(it)
    assert 11 == next(it)
    assert 10 == next(it)
    assert 11 == next(it)
    with pytest.raises(StopIteration):
        next(it)

@given(st.integers())
def test_repeatfunc(x):
    def f():
        return "test"

    # unlimited count
    it = iters.repeatfunc(f)
    assert "test" == next(it)
    assert "test" == next(it)
    assert "test" == next(it)

    # limited
    it = iters.repeatfunc(f, 2)
    assert "test" == next(it)
    assert "test" == next(it)
    with pytest.raises(StopIteration):
        next(it)


def test_grouper():
    # without fill value (default should be None)
    a, b, c = iters.grouper(3, "ABCDEFG")
    assert ["A", "B", "C"] == list(a)
    assert ["D", "E", "F"] == list(b)
    assert ["G", None, None] == list(c)

    # with fill value
    a, b, c = iters.grouper(3, "ABCDEFG", "x")
    assert ["A", "B", "C"] == list(a)
    assert ["D", "E", "F"] == list(b)
    assert ["G", "x", "x"] == list(c)


def test_roundrobin():
    r = iters.roundrobin('ABC', 'D', 'EF')
    assert ["A", "D", "E", "B", "F", "C"] == list(r)


def test_partition():
    def is_odd(x):
        return x % 2 == 1

    before, after = iters.partition(is_odd, range(5))
    assert [0, 2, 4] == list(before)
    assert [1, 3] == list(after)


def test_splitat():
    before, after = iters.splitat(2, range(5))
    assert [0, 1] == list(before)
    assert [2, 3, 4] == list(after)


def test_splitby():
    def is_even(x):
        return x % 2 == 0

    before, after = iters.splitby(is_even, range(5))
    assert [0] == list(before)
    assert [1, 2, 3, 4] == list(after)


def test_powerset():
    ps = iters.powerset([1, 2])
    assert [tuple(), (1,), (2,), (1, 2)] == list(ps)


def test_pairwise():
    ps = iters.pairwise([1, 2, 3, 4])
    assert [(1, 2), (2, 3), (3, 4)] == list(ps)


def test_iter_except():
    d = ["a", "b", "c"]
    it = iters.iter_suppress(d.pop, IndexError)
    assert ["c", "b", "a"] == list(it)


def test_flatten():
    # flatten nested lists
    assert [1, 2, 3, 4] == list(iters.flatten([[1, 2], [3, 4]]))
    assert [1, 2, 3, 4, 5, 6] == list(iters.flatten([[1, 2], [3, [4, 5, 6]]]))
    # flatten nested tuples, sets, and frozen sets
    assert [1, 2, 3, 4, 5, 6] == list(iters.flatten(((1, 2), (3, (4, 5, 6)))))
    assert [1, 2, 3] == list(iters.flatten(set([1, frozenset([2, 3])])))
    # flatten nested generators
    generators = ((num + 1 for num in range(0, n)) for n in range(1, 4))
    assert [1, 1, 2, 1, 2, 3] == list(iters.flatten(generators))
    # flat list should return it
    assert [1, 2, 3] == list(iters.flatten([1, 2, 3]))
    # Don't flatten strings, bytes, or bytearrays
    assert [2, "abc", 1] == list(iters.flatten([2, "abc", 1]))
    assert [2, b'abc', 1] == list(iters.flatten([2, b'abc', 1]))
    assert [2, bytearray(b'abc'), 1] == list(iters.flatten([2, bytearray(b'abc'), 1]))


def test_accumulate():
    assert [1, 3, 6, 10, 15] == list(iters.accumulate([1, 2, 3, 4, 5]))
    assert [1, 2, 6, 24, 120] == list(iters.accumulate([1, 2, 3, 4, 5], operator.mul))


def test_filterfalse():
    l = iters.filterfalse(lambda x: x > 10, [1, 2, 3, 11, 12])
    assert [1, 2, 3] == list(l)


def test_iterate():
    f = F() << _ * 2
    it = iters.iterate(f, 2)
    assert 2 == next(it)
    assert 4 == next(it)
    assert 8 == next(it)
    assert 16 == next(it)



import collections

from ..interface import Contract, ContractNotRespected
from ..syntax import (add_contract, W, contract_expression, O, S, add_keyword,
    Keyword)


class Collection(Contract):

    def __init__(self, length_contract=None,
                    elements_contract=None, where=None):
        Contract.__init__(self, where)
        self.length_contract = length_contract
        self.elements_contract = elements_contract

        try:
            # latest python 3.6+
            self.__collection_types = collections.abc.Collection
        except:
            try:
                # older python 3
                self.__collection_types = collections.Collection
            except:
                # python 2
                self.__collection_types = (collections.Sequence, collections.Set, collections.Mapping, collections.deque,)

    def check_contract(self, context, value, silent):
        if not isinstance(value, self.__collection_types):
            error = 'Expected a Collection, got %r.' % value.__class__.__name__
            raise ContractNotRespected(self, error, value, context)

        if self.length_contract is not None:
            self.length_contract._check_contract(context, len(value), silent)

        if self.elements_contract is not None:
            for element in value:
                self.elements_contract._check_contract(context, element, silent)

    def __str__(self):
        s = 'collection'
        if self.length_contract is not None:
            s += '[%s]' % self.length_contract
        if self.elements_contract is not None:
            s += '(%s)' % self.elements_contract
        return s

    def __repr__(self):
        return 'Collection({0!r},{0!r})'.format(self.length_contract, self.elements_contract)

    @staticmethod
    def parse_action(s, loc, tokens):
        where = W(s, loc)
        length_contract = tokens.get('length_contract', None)
        elements_contract = tokens.get('elements_contract', None)
        return Collection(length_contract, elements_contract, where=where)


list_contract = (Keyword('collection') -
                 O(S('[') - contract_expression('length_contract') - S(']')) +
                 O(S('(') - contract_expression('elements_contract') - S(')')))
list_contract.setParseAction(Collection.parse_action)

list_contract.setName('Collection contract')
add_keyword('collection')
add_contract(list_contract)



from . import good, fail

good('Container', []) 
fail('Container', 1)
good('Hashable', 1)
# counterexample? 
good('Iterable', []) 
fail('Iterable', 1)
good('Iterable', {})
good('Iterator', [].__iter__())
fail('Iterator', []) 
good('Sized', [])
good('Sized', {})
fail('Sized', lambda: None)
good('Sized', '')
good('Callable', lambda: None)
fail('Callable', [])
good('Sequence', [])
good('Sequence', (1,))
good('Sequence', '')
fail('Sequence', {})
good('Set', set([]))
fail('Set', [])
good('MutableSequence', [])
fail('MutableSequence', (1,))
good('MutableSet', set([]))
fail('MutableSet', frozenset([]))
good('Mapping', {})
fail('Mapping', [])
good('MutableMapping', {})
#good('MappingView', {}.keys())
#good('ItemsView', {}.items())
#good('ValuesView', {}.values())
